/*
 * The TiLib: wavelet based lossy image compression library
 * Copyright (C) 1998-2004 Alexander Simakov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * QuikInfo:
 *
 * Simple example of using TiLib for image coding (see readme.txt).
 * 
 * E-mail: xander@entropyware.info
 * Internet: http://www.entropyware.info
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "pbm.h"
#include "tilib.h"
#include "errcodes.h"

#define MAX_LINE 1024

int encode;
char infile[MAX_LINE];  /* input file name */
char outfile[MAX_LINE]; /* output file name */
int levels;             /* number of transform levels */
int lum;                /* see readme.txt */
int cb;                 /* see readme.txt */
int cr;                 /* see readme.txt */
int size;               /* desired encoded image size */
int filter;             /* use Butterworth or Daubechies filter */

void usage()
{
  printf("TICODEC - TiLib ver 0.9.1 sample codec.\n");
  printf("Copyright (C) 1998-2004 Alexander Simakov.\n");

  printf("\nusage:  ticodec [-e|-d] [-i:infile] [-o:outfile]\n\
                [-l:levels] [-s:size]\n\
                [-y:Y%% -b:Cb%% -r:Cr%%]\n\
                [-B|-D]\n");

  printf("\noptions:\n\n");

  printf("-e encode image | -l number of transform levels, i.e. scales      (require -e)\n");
  printf("-d decode image | -s desired encoded image size in bytes          (require -e)\n");
  printf("-i input file   | -y, -b, -r mix Y, Cb, Cr streams in this ratio  (require -e)\n");
  printf("-o output file  | -B, -D use (B)utterworth or (D)aubechies filter (require -e)\n");

  printf("\nexamples:\n\n");

  printf("ticodec -e -i:foto.ppm -o:foto.Ti -s:20000\n");
  printf("ticodec -e -i:foobar.pgm -o:foobar.Ti -s:100 -D\n");
  printf("ticodec -e -i:test.pgm -o:test.Ti -l:5 -s:777\n");
  printf("ticodec -e -i:simakov.ppm -o:simakov.Ti -l:8 -s:65536 -y:90 -b:5 -r:5 -B\n");
  printf("ticodec -d -i:foto.Ti -o:foto-out.ppm\n");

  exit(1);
}

void validate_args(int argc, char **argv)
{
  int ed_flg, i_flg, o_flg, l_flg, s_flg, y_flg, b_flg, r_flg, BD_flg;
  int i;

  ed_flg = i_flg = o_flg = l_flg = s_flg = y_flg = b_flg = r_flg = BD_flg = 0;

  for (i = 1; i < argc; i++) {

    if ((argv[i][0] == '-') || (argv[i][0] == '/')) {

      switch (argv[i][1]) {

        case 'e':
          if (ed_flg) usage();
          ed_flg = 1; encode = 1;
          break;

        case 'd':
          if (ed_flg) usage();
          ed_flg = 1; encode = 0;
          break;

        case 'i':
          if (strlen(argv[i]) <= 3 || i_flg) usage();
          strncpy(infile, &argv[i][3], MAX_LINE - 1);
          infile[MAX_LINE - 1] = 0;
          i_flg = 1;
          break;

        case 'o':
          if (strlen(argv[i]) <= 3 || o_flg) usage();
          strncpy(outfile, &argv[i][3], MAX_LINE - 1);
          outfile[MAX_LINE - 1] = 0;
          o_flg = 1;
          break;

        case 'l':
          if (strlen(argv[i]) <= 3 || l_flg) usage();
          levels = atoi(&argv[i][3]);
          l_flg = 1;
          break;

        case 's':
          if (strlen(argv[i]) <= 3 || s_flg) usage();
          size = atoi(&argv[i][3]);
          s_flg = 1;
          break;

        case 'y':
          if (strlen(argv[i]) <= 3 || y_flg) usage();
          lum = atoi(&argv[i][3]);
          y_flg = 1;
          break;

        case 'b':
          if (strlen(argv[i]) <= 3 || b_flg) usage();
          cb = atoi(&argv[i][3]);
          b_flg = 1;
          break;

        case 'r':
          if (strlen(argv[i]) <= 3 || r_flg) usage();
          cr = atoi(&argv[i][3]);
          r_flg = 1;
          break;

        case 'B':
          if (BD_flg) usage();
          filter = BUTTERWORTH;
          BD_flg = 1;
          break;

        case 'D':
          if (BD_flg) usage();
          filter = DAUB97;
          BD_flg = 1;
          break;

        default:
          usage();
          break;
      }

    }

  }

  /* check options */

  if (ed_flg == 0 || i_flg == 0 || o_flg == 0) usage();

  if (encode == 1) {
    if (s_flg == 0) usage();
    if (size < 26) usage(); /* HDRSIZE + 2 + 2 + 2 */
    if ((y_flg + b_flg + r_flg) != 0 && (y_flg + b_flg + r_flg) != 3) usage();
    if (l_flg == 0) levels = 0;
    if (BD_flg == 0) filter = DAUB97;
    if (y_flg + b_flg + r_flg == 0) lum = cb = cr = 0;
    if ((y_flg + b_flg + r_flg == 3) && (lum == 0 || cb == 0 || cr == 0)) usage();
    if ((y_flg + b_flg + r_flg == 3) && (lum + cb + cr != 100)) usage();
  } else {
    if (l_flg + s_flg + y_flg + b_flg + r_flg + BD_flg != 0) usage();
  }
}

void CompressFile()
{
  FILE *in_file, *out_file;
  unsigned char *in_buf, *out_buf;
  int width, height, actual_size;
  int result;
  pbm_hdr h;

  in_buf = out_buf = NULL;

  /* open files */

  in_file = fopen(infile, "rb");
  out_file = fopen(outfile, "wb");

  if (in_file == NULL || out_file == NULL) {
    perror("fopen() failed");
    exit(1);
  }

  /* read image header */

  read_hdr(in_file, &h);

  width = h.width;
  height = h.height;

  /* allocate memory */

  in_buf = (unsigned char *) malloc(width * height * (h.type == PGM ? 1 : 3));
  out_buf = (unsigned char *) malloc(size);

  if (in_buf == NULL || out_buf == NULL) {
    perror("malloc() failed");
    exit(1);
  }

  /* read whole image into the buffer */

  fread(in_buf, 1, width * height * (h.type == PGM ? 1 : 3), in_file);

  /* compress it (with only one function call!) */

  result = TiCompress(in_buf, out_buf, width, height, filter, (h.type == PGM ? GRAYSCALE : TRUECOLOR),
  size, &actual_size, lum, cb, cr, levels);

  /* if something wrong ... */

  if (result != OK) {
    printf("TiCompress() failed: %d\n", result);
    exit(1);
  }

  /* store encoded stream to disk */

  fwrite(out_buf, 1, actual_size, out_file);
}

void DecompressFile()
{
  FILE *in_file, *out_file;
  unsigned char *in_buf, *out_buf;
  int width, height, type, result, stream_size;

  in_buf = out_buf = NULL;

  /* open files */

  in_file = fopen(infile, "rb");
  out_file = fopen(outfile, "wb");

  if (in_file == NULL || out_file == NULL) {
    perror("fopen() failed");
    exit(1);
  }

  stream_size = file_size(in_file);

  /* allocate memory for a whole file */

  in_buf = (unsigned char *) malloc(stream_size);

  if (in_buf == NULL) {
    perror("malloc() failed");
    exit(1);
  }

  /* read it */

  fread(in_buf, 1, stream_size, in_file);

  /* check .Ti image header */

  result = TiCheckHeader(in_buf, &width, &height, &type, stream_size);

  /* ... damaged :( */

  if (result != OK) {
    printf("TiCheckHeader() failed: %d\n", result);
    exit(1);
  }

  /* allocate memory for a decoded image */

  out_buf = (unsigned char *) malloc(width * height * (type == GRAYSCALE ? 1 : 3));

  /* decode it! */

  result = TiDecompress(in_buf, out_buf, width, height, type, stream_size);

  /* if something wrong ... */

  if (result != OK) {
    printf("TiDecompress() failed: %d\n", result);
    exit(1);
  }

  /* write PGM of PPM header */

  fprintf(out_file, "P%s\n%d %d\n255\n", type == GRAYSCALE ? "5" : "6", width, height);

  /* store decoded image (PGM or PPM) to disk */

  fwrite(out_buf, 1, width * height * (type == GRAYSCALE ? 1 : 3), out_file);
}

int main(int argc, char *argv[])
{
  validate_args(argc, argv);

  if (encode) CompressFile();
  else DecompressFile();

  return 0;
}
